<?php

abstract class FastCommentsIntegrationCore {
    public $baseUrl;
    public $integrationType;

    public function __construct($integrationType, $host = 'https://fastcomments.com') {
        if (!$integrationType) {
            throw new RuntimeException('An integration type is required! Ex: new FastCommentsIntegrationCore("wordpress")');
        }
        if (!$host) {
            throw new RuntimeException('An integration host is required! A valid default is available, did you try to set this to a weird value?');
        }
        $this->integrationType = $integrationType;
        $this->baseUrl = "$host/integrations/v1";
    }

    public abstract function activate();

    public abstract function deactivate();

    public abstract function createUUID();

    public abstract function getDomain();

    public abstract function getSettingValue($settingName);

    public abstract function setSettingValue($settingName, $settingValue);

    public abstract function makeHTTPRequest($method, $url, $body);

    public abstract function getCurrentUser();

    public abstract function getLoginURL();

    public abstract function getLogoutURL();

    public abstract function handleEvents($events);

    public abstract function getCommentCount($afterId);

    public abstract function getComments($afterId);

    public function base64Encode($stringValue) {
        return base64_encode($stringValue);
    }

    public function log($level, $message) {
        switch ($level) {
            case 'debug':
                echo "DEBUG:::" . $message . "\n";
                break;
            case 'info':
                echo "INFO:::" . $message . "\n";
                break;
            case 'error':
                echo "ERROR:::" . $message . "\n";
                break;
        }
    }

    public function getConfig($timestamp, $urlId, $url, $isReadonly) {
        $ssoKey = $this->getSettingValue('fastcomments_sso_key');
        $tenantId = $this->getSettingValue('fastcomments_tenant_id');
        $isSSOEnabled = $ssoKey && $this->getSettingValue('fastcomments_sso_enabled');
        $result = new stdClass();
        $result->tenantId = ($tenantId) ? $tenantId : 'demo';
        $result->urlId = $urlId;
        $result->url = $url;
        $result->readonly = $isReadonly;
        $result->sso = ($isSSOEnabled) ? $this->getSSOConfig($timestamp, $ssoKey) : null;
        return $result;
    }

    public function getSSOConfig($timestamp, $ssoKey) {
        $result = new stdClass();
        $result->timestamp = $timestamp;
        $sso_user = new stdClass();
        $user = $this->getCurrentUser();
        if ($user) {
            $sso_user->id = $user->id;
            $sso_user->email = $user->email;
            $sso_user->username = $user->username;
            $sso_user->avatar = $user->avatar;
            $sso_user->optedInNotifications = true;
        }
        $userDataJSONBase64 = $this->base64Encode(json_encode($sso_user));
        $verificationHash = hash_hmac('sha256', $timestamp->userDataJSONBase64, $ssoKey);
        $result->userDataJSONBase64 = $userDataJSONBase64;
        $result->verificationHash = $verificationHash;
        $result->loginURL = $this->getLoginURL();
        $result->logoutURL = $this->getLogoutURL();
        return $result;
    }

    public function enableSSO() {
        $token = $this->getSettingValue("fastcomments_token");
        $apiSecretResponseRaw = $this->makeHTTPRequest('GET', "$this->baseUrl/api-secret?token=$token", null);
        $this->log('debug', 'Secret Token Fetch Response Code:' . $apiSecretResponseRaw->responseStatusCode);
        $apiSecretResponse = json_decode($apiSecretResponseRaw->responseBody);
        if ($apiSecretResponse->status === 'success' && $apiSecretResponse->secret) {
            $this->setSettingValue('fastcomments_sso_key', $apiSecretResponse->secret);
            $this->setSettingValue('fastcomments_sso_enabled', true);
        } else {
            $this->log('error', $apiSecretResponse->reason);
            throw new RuntimeException("API did not return success response when trying to get the key!");
        }
    }

    public function disableSSO() {
        $this->setSettingValue('fastcomments_sso_key', null);
        $this->setSettingValue('fastcomments_sso_enabled', false);
    }

    public function tick() {
        $this->log('debug', "BEGIN Tick");
        $nextStateMachineName = 'integrationStateInitial';
        while ($nextStateMachineName) {
            $this->log('debug', 'Next state machine:' . $nextStateMachineName);
            $nextStateMachineName = call_user_func(array($this, $nextStateMachineName));
        }
        $this->log('debug', "END Tick");
    }

    public function integrationStateInitial() {
        $tenantId = $this->getSettingValue('fastcomments_tenant_id');
        $token = $this->getSettingValue('fastcomments_token');
        $isTokenValidated = $this->getSettingValue('fastcomments_token_validated');
        if ($tenantId && $token && $isTokenValidated) {
            return 'integrationStatePollNext';
        } else {
            if ($token) {
                return 'integrationStateValidateToken';
            } else {
                return 'integrationStateCreateToken';
            }
        }
    }

    public function integrationStateValidateToken() {
        $token = $this->getSettingValue('fastcomments_token');
        if ($token) {
            $domainName = $this->getDomain();
            $rawTokenUpsertResponse = $this->makeHTTPRequest('PUT', "$this->baseUrl/token?token=$token&integrationType=$this->integrationType&domain=$domainName", null);
            $tokenUpsertResponse = json_decode($rawTokenUpsertResponse->responseBody);
            if ($tokenUpsertResponse->status === 'success' && $tokenUpsertResponse->isTokenValidated === true) {
                $this->setSettingValue('fastcomments_tenant_id', $tokenUpsertResponse->tenantId);
                $this->setSettingValue('fastcomments_token_validated', true);
            }
            return null;
        } else {
            return 'integrationStateCreateToken';
        }
    }

    public function integrationStateCreateToken() {
        $newUUID = $this->createUUID();
        $this->setSettingValue('fastcomments_token', $newUUID);
        return null;
    }

    public function integrationStatePollNext() {
        // One idea to consider, that'd be easier to understand, would be to store the commands locally and a queue and process them.
        // This removes the weird logic where each time a command is finished processing, we advance the fastcomments_stream_last_fetch_timestamp.
        // The reason this logic is weird is the two things are relatively far from each other, potentially being bug prone.
        $token = $this->getSettingValue('fastcomments_token');
        if ($token) {
            $lastFetchDate = $this->getSettingValue('fastcomments_stream_last_fetch_timestamp');
            $lastFetchDateToSend = $lastFetchDate ? $lastFetchDate : 0;
            $rawIntegrationStreamResponse = $this->makeHTTPRequest('GET', "$this->baseUrl/commands?token=$token&fromDateTime=$lastFetchDateToSend", null);
            $this->log('debug', 'Stream response status: ' . $rawIntegrationStreamResponse->responseStatusCode);
            if ($rawIntegrationStreamResponse->responseStatusCode === 200) {
                $response = json_decode($rawIntegrationStreamResponse->responseBody);
                if ($response->status === 'success' && $response->commands) {
                    foreach ($response->commands as $command) {
                        $this->log('debug', "Processing command $command->command");
                        switch ($command->command) {
                            case 'FetchEvents':
                                $this->commandFetchEvents($token);
                                break;
                            case 'SendComments':
                                $this->commandSendComments($token);
                                break;
                        }
                    }
                }
            }
        } else {
            $this->log('error', "Cannot fetch commands, fastcomments_token not set.");
        }
        return null;
    }

    public function commandFetchEvents($token) {
        $fromDateTime = $this->getSettingValue('fastcomments_stream_last_fetch_timestamp');
        $hasMore = true;
        $startedAt = time();
        while ($hasMore && time() - $startedAt < 30 * 1000) {
            $this->log('debug', 'Send events command loop...');
            $fromDateTimeToSend = $fromDateTime ? $fromDateTime : 0;
            $rawIntegrationEventsResponse = $this->makeHTTPRequest('GET', "$this->baseUrl/events?token=$token&fromDateTime=$fromDateTimeToSend", null);
            $response = json_decode($rawIntegrationEventsResponse->responseBody);
            if ($response->status === 'success') {
                $count = count($response->events);
                $this->log('info', "Got events count=[$count]");
                if ($response->events && count($response->events) > 0) {
                    $this->handleEvents($response->events);
                    $fromDateTime = strtotime($response->events[count($response->events) - 1]->createdAt);
                }
                $hasMore = $response->hasMore;
                $this->setSettingValue('fastcomments_stream_last_fetch_timestamp', $fromDateTime);
            } else {
                $this->log('error', "Failed to get events: {$rawIntegrationEventsResponse}");
                break;
            }
        }
    }

    private function canAckLock($name, $windowSeconds) {
        $settingName = $this->getLockName($name);
        $lastTime = $this->getSettingValue($settingName);
        $now = time();
        if ($lastTime && $now - $lastTime < $windowSeconds) {
            return false;
        }
        $this->setSettingValue($settingName, $now);
        return true;
    }

    private function getLockName($name) {
        return "lock_$name";
    }

    private function clearLock($name) {
        $settingName = $this->getLockName($name);
        $this->setSettingValue($settingName, null);
    }

    public function commandSendComments($token) {
        /**
         * Fetch 500 comments a time from the DB.
         * Split them up into chunks of 100.
         * If the server complains the payload is too large, recursively split the chunk by / 10.
         */
        $this->log('debug', 'Starting to send comments');
        if (!$this->canAckLock("commandSendComments", 60)) {
            $this->log('debug', 'Can not send right now, waiting for previous attempt to finish.');
            return;
        }
        $lastSendDate = $this->getSettingValue('fastcomments_stream_last_send_timestamp');
        $lastSentId = $this->getSettingValue('fastcomments_stream_last_send_id');
        $commentCount = $this->getCommentCount($lastSentId ? $lastSentId : 0);
        if ($commentCount == 0) {
            $this->log('debug', "No comments to send. Telling server. lastSendDate=[$lastSendDate] lastSentId=[$lastSentId]");
            $requestBody = json_encode(
                array(
                    "countRemaining" => 0,
                    "comments" => array()
                )
            );
            $httpResponse = $this->makeHTTPRequest('POST', "$this->baseUrl/comments?token=$token", $requestBody);
            $this->log('debug', "Got POST /comments response status code=[$httpResponse->responseStatusCode]");
            $this->setSetupDone();
            return;
        }
        $this->log('debug', 'Send comments command loop...');
        $getCommentsResponse = $this->getComments($lastSentId ? $lastSentId : 0);
        if ($getCommentsResponse['status'] === 'success') {
            $count = count($getCommentsResponse['comments']);
            $this->log('info', "Got comments to send count=[$count] from totalCount=[$commentCount] lastSendDate=[$lastSendDate] lastSentId=[$lastSentId]");
            $countRemaining = $commentCount;
            $chunkSize = 100;
            if ($countRemaining > 0) {
                $commentChunks = array_chunk($getCommentsResponse['comments'], $chunkSize);
                foreach ($commentChunks as $chunk) {
                    // for this chunk, attempt to send the whole thing. If it fails, split it up.
                    $chunkAttemptsRemaining = 5;
                    $dynamicChunkSize = $chunkSize;
                    $dynamicChunks = array($chunk);
                    while ($chunkAttemptsRemaining > 0) {
                        foreach ($dynamicChunks as $dynamicChunk) {
                            $lastComment = $dynamicChunk[count($dynamicChunk) - 1];
                            $lastCommentFromDateTime = strtotime($lastComment['date']) * 1000;
                            $countRemainingIfSuccessful = $countRemaining - count($dynamicChunk);
                            $requestBody = json_encode(
                                array(
                                    "countRemaining" => $countRemainingIfSuccessful,
                                    "comments" => $dynamicChunk
                                )
                            );
                            $httpResponse = $this->makeHTTPRequest('POST', "$this->baseUrl/comments?token=$token", $requestBody);
                            $this->log('debug', "Got POST /comments response status code=[$httpResponse->responseStatusCode] and chunk size $dynamicChunkSize");
                            if ($httpResponse->responseStatusCode === 200) {
                                $response = json_decode($httpResponse->responseBody);
                                if ($response->status === 'success') {
                                    $countRemaining = $countRemainingIfSuccessful;
                                    $fromDateTime = $lastCommentFromDateTime;
                                    $this->setSettingValue('fastcomments_stream_last_send_timestamp', $fromDateTime);
                                    $this->setSettingValue('fastcomments_stream_last_send_id', $lastComment['externalId']);
                                    if ($countRemaining <= 0) {
                                        $this->setSetupDone();
                                    }
                                }
                                $chunkAttemptsRemaining = 0; // done
                            } else if ($httpResponse->responseStatusCode === 413 && $dynamicChunkSize > 1) {
                                $this->log('debug', "$dynamicChunkSize too big, splitting.");
                                $dynamicChunks = array_chunk($chunk, max((int)($dynamicChunkSize / 10), 1));
                                break; // break out of the dynamic chunks loop and run it again
                            }
                        }
                        $chunkAttemptsRemaining--;
                    }
                }
            } else {
                $this->setSetupDone();
            }
        } else {
            $status = $getCommentsResponse['status'];
            $comments = $getCommentsResponse['comments'];
            $this->log('error', "Failed to get comments to send: status=[$status] comments=[$comments]");
        }
        $this->clearLock("commandSendComments");
        $this->log('debug', 'Done sending comments');
    }

    private function setSetupDone() {
        $this->setSettingValue('fastcomments_setup', true);
        $this->setSettingValue('fastcomments_stream_last_fetch_timestamp', time() * 1000);
        $this->clearLock("commandSendComments");
    }

}
